import pandas as pd

# make_unique_labels

def main(net, df=None):
  '''
  Run in load_data module (which runs when file is loaded or dataframe is loaded),
  check for duplicate row/col names, and add index to names if necesary
  '''
  if df is None:
    df = net.export_df()

  # rows
  #############
  rows = df.index.tolist()
  if type(rows[0]) is str:

    if len(rows) != len(list(set(rows))):
      print('warning: making row names unique')
      new_rows = add_index_list(rows)
      df.index = new_rows

  elif type(rows[0]) is tuple:

    row_names = []
    for inst_row in rows:
      row_names.append(inst_row[0])

    if len(row_names) != len(list(set(row_names))):
      print('warning: making row names unique')
      row_names = add_index_list(row_names)

      # add back to tuple
      new_rows = []
      for inst_index in range(len(rows)):
        inst_row = rows[inst_index]
        new_row = list(inst_row)
        new_row[0] = row_names[inst_index]
        new_row = tuple(new_row)
        new_rows.append(new_row)

      df.index = new_rows

  # cols
  #############
  cols = df.columns.tolist()
  if type(cols[0]) is str:

    # list column names
    if len(cols) != len(list(set(cols))):
      print('warning: making col names unique')
      new_cols = add_index_list(cols)
      df.columns = new_cols

  elif type(cols[0]) is tuple:

    col_names = []
    for inst_col in cols:
      col_names.append(inst_col[0])

    if len(col_names) != len(list(set(col_names))):
      print('warning: making col names unique')
      col_names = add_index_list(col_names)

      # add back to tuple
      new_cols = []
      for inst_index in range(len(cols)):
        inst_col = cols[inst_index]
        new_col = list(inst_col)
        new_col[0] = col_names[inst_index]
        new_col = tuple(new_col)
        new_cols.append(new_col)

      df.columns = new_cols

  # return dataframe with unique names
  return df

def add_index_list(nodes):

  new_nodes = []
  for i in range(len(nodes)):
    index = i + 1
    inst_node = nodes[i]
    new_node = inst_node + '-' + str(index)
    new_nodes.append(new_node)

  return new_nodes



def check_categories(lines):
  '''
  find out how many row and col categories are available
  when loading data from file
  '''
  # count the number of row categories
  rcat_line = lines[0].split('\t')

  # calc the number of row names and categories
  num_rc = 0
  found_end = False

  # skip first tab
  for inst_string in rcat_line[1:]:
    if inst_string == '':
      if found_end is False:
        num_rc = num_rc + 1
    else:
      found_end = True

  max_rcat = 15
  if max_rcat > len(lines):
    max_rcat = len(lines) - 1

  num_cc = 0
  for i in range(max_rcat):
    ccat_line = lines[i + 1].split('\t')

    # make sure that line has length greater than one to prevent false cats from
    # trailing new lines at end of matrix
    if ccat_line[0] == '' and len(ccat_line) > 1:
      num_cc = num_cc + 1

  num_labels = {}
  num_labels['row'] = num_rc + 1
  num_labels['col'] = num_cc + 1

  return num_labels

def dict_cat(net, define_cat_colors=False):
  '''
  make a dictionary of node-category associations
  '''
  net.persistent_cat_colors = True

  for inst_rc in ['row', 'col']:
    inst_keys = list(net.dat['node_info'][inst_rc].keys())
    all_cats = [x for x in inst_keys if 'cat-' in x]

    for inst_name_cat in all_cats:

      dict_cat = {}
      tmp_cats = net.dat['node_info'][inst_rc][inst_name_cat]
      tmp_nodes = net.dat['nodes'][inst_rc]

      for i in range(len(tmp_cats)):
        inst_cat = tmp_cats[i]
        inst_node = tmp_nodes[i]

        if inst_cat not in dict_cat:
          dict_cat[inst_cat] = []

        dict_cat[inst_cat].append(inst_node)

      tmp_name = 'dict_' + inst_name_cat.replace('-', '_')
      net.dat['node_info'][inst_rc][tmp_name] = dict_cat

  # merge with old cat_colors by default
  cat_colors = net.viz['cat_colors']

  if define_cat_colors == True:
    cat_number = 0

    for inst_rc in ['row', 'col']:

      inst_keys = list(net.dat['node_info'][inst_rc].keys())
      all_cats = [x for x in inst_keys if 'cat-' in x]

      for cat_index in all_cats:

        if cat_index not in cat_colors[inst_rc]:
          cat_colors[inst_rc][cat_index] = {}

        cat_names = sorted(list(set(net.dat['node_info'][inst_rc][cat_index])))

        # loop through each category name and assign a color
        for tmp_name in cat_names:

          # using the same rules as the front-end to define cat_colors
          inst_color = get_cat_color(cat_number + cat_names.index(tmp_name))

          check_name = tmp_name

          # check if category is string type and non-numeric
          try:
            float(check_name)
            is_string_cat = False
          except:
            is_string_cat = True

          if is_string_cat == True:
            # check for default non-color
            if ': ' in check_name:
              check_name = check_name.split(': ')[1]

            # if check_name == 'False' or check_name == 'false':
            if 'False' in check_name or 'false' in check_name:
              inst_color = '#eee'

            if 'Not ' in check_name:
              inst_color = '#eee'

          # do not overwrite old colors
          if tmp_name not in cat_colors[inst_rc][cat_index] and is_string_cat:

            cat_colors[inst_rc][cat_index][tmp_name] = inst_color
            # print('overwrite: ' + tmp_name + ' -> ' + str(inst_color))

          cat_number = cat_number + 1

    net.viz['cat_colors'] = cat_colors

    net.viz['global_cat_colors'] = {}

    global_cat_colors = {}
    for axis in net.viz['cat_colors']:
        for cat_index in net.viz['cat_colors'][axis]:

            inst_dict = net.viz['cat_colors'][axis][cat_index]

            for inst_full_name in inst_dict:

                inst_name = inst_full_name.split(': ')[1]
                inst_color = inst_dict[inst_full_name]

                global_cat_colors[inst_name] = inst_color

    net.viz['global_cat_colors'] = global_cat_colors


def calc_cat_clust_order(net, inst_rc):
  '''
  cluster category subset of data
  '''
  from .__init__ import Network
  from copy import deepcopy
  from . import calc_clust, run_filter

  inst_keys = list(net.dat['node_info'][inst_rc].keys())
  all_cats = [x for x in inst_keys if 'cat-' in x]

  if len(all_cats) > 0:

    for inst_name_cat in all_cats:

      tmp_name = 'dict_' + inst_name_cat.replace('-', '_')
      dict_cat = net.dat['node_info'][inst_rc][tmp_name]

      unordered_cats = dict_cat.keys()

      ordered_cats = order_categories(unordered_cats)

      # this is the ordering of the columns based on their category, not
      # including their clustering ordering within category
      all_cat_orders = []
      tmp_names_list = []
      for inst_cat in ordered_cats:

        inst_nodes = dict_cat[inst_cat]

        tmp_names_list.extend(inst_nodes)

      names_clust_list = tmp_names_list

      # calc category-cluster order
      final_order = []

      for i in range(len(net.dat['nodes'][inst_rc])):

        inst_node_name = net.dat['nodes'][inst_rc][i]
        inst_node_num = names_clust_list.index(inst_node_name)

        final_order.append(inst_node_num)

      inst_index_cat = inst_name_cat.replace('-', '_') + '_index'

      net.dat['node_info'][inst_rc][inst_index_cat] = final_order


def order_categories(unordered_cats):
  '''
  If categories are strings, then simple ordering is fine.
  If categories are values then I'll need to order based on their values.
  The final ordering is given as the original categories (including titles) in a
  ordered list.
  '''

  no_titles = remove_titles(unordered_cats)

  all_are_numbers = check_all_numbers(no_titles)

  if all_are_numbers:
    ordered_cats = order_cats_based_on_values(unordered_cats, no_titles)
  else:
    ordered_cats = sorted(unordered_cats)

  return ordered_cats


def order_cats_based_on_values(unordered_cats, values_list):
  import pandas as pd

  try:
    # convert values_list to values
    values_list = [float(i) for i in values_list]

    inst_series = pd.Series(data=values_list, index=unordered_cats)

    inst_series.sort_values(inplace=True)

    ordered_cats = inst_series.index.tolist()

    # ordered_cats = unordered_cats
  except:
    # keep default ordering if error occurs
    print('error sorting cats based on values ')
    ordered_cats = unordered_cats

  return ordered_cats

def check_all_numbers(no_titles):
  all_numbers = True
  for tmp in no_titles:
    if is_number(tmp) == False:
      all_numbers = False

  return all_numbers

def remove_titles(cats):
  from copy import deepcopy

  # check if all have titles
  ###########################
  all_have_titles = True

  for inst_cat in cats:
    if is_number(inst_cat) == False:
      if ': ' not in inst_cat:
        all_have_titles = False
    else:
      all_have_titles = False

  if all_have_titles:
    no_titles = cats
    no_titles = [i.split(': ')[1] for i in no_titles]

  else:
    no_titles = cats

  return no_titles

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def get_cat_color(cat_num):

  all_colors = [ "#393b79", "#aec7e8", "#ff7f0e", "#ffbb78", "#98df8a", "#bcbd22",
    "#404040", "#ff9896", "#c5b0d5", "#8c564b", "#1f77b4", "#5254a3", "#FFDB58",
    "#c49c94", "#e377c2", "#7f7f7f", "#2ca02c", "#9467bd", "#dbdb8d", "#17becf",
    "#637939", "#6b6ecf", "#9c9ede", "#d62728", "#8ca252", "#8c6d31", "#bd9e39",
    "#e7cb94", "#843c39", "#ad494a", "#d6616b", "#7b4173", "#a55194", "#ce6dbd",
    "#de9ed6"];

  inst_color = all_colors[cat_num % len(all_colors)]

  return inst_color

def add_cats(net, axis, cat_data):

  try:
    df = net.export_df()

    if axis == 'row':
      labels = df.index.tolist()
    elif axis == 'col':
      labels = df.columns.tolist()

    if 'title' in cat_data:
      inst_title = cat_data['title']
    else:
      inst_title = 'New Category'

    all_cats = cat_data['cats']

    # loop through all labels
    new_labels = []
    for inst_label in labels:

      if type(inst_label) is tuple:
        check_name = inst_label[0]
        found_tuple = True
      else:
        check_name = inst_label
        found_tuple = False

      if ': ' in check_name:
        check_name = check_name.split(': ')[1]

      # default to False for found cat, overwrite if necessary
      found_cat = inst_title + ': False'

      # check all categories in cats
      for inst_cat in all_cats:

        inst_names = all_cats[inst_cat]

        if check_name in inst_names:
          found_cat = inst_title + ': ' + inst_cat

      # add category to label
      if found_tuple is True:
        new_label = inst_label + (found_cat,)
      else:
        new_label = (inst_label, found_cat)

      new_labels.append(new_label)


    # add labels back to DataFrame
    if axis == 'row':
      df.index = new_labels
    elif axis == 'col':
      df.columns = new_labels

    net.load_df(df)

  except:
    print('error adding new categories')







def df_filter_row_sum(df, threshold, take_abs=True):
  ''' filter rows in matrix at some threshold
  and remove columns that have a sum below this threshold '''

  from copy import deepcopy
  from .__init__ import Network
  net = Network()

  if take_abs is True:
    df_copy = deepcopy(df.abs())
  else:
    df_copy = deepcopy(df)

  ini_rows = df_copy.index.values.tolist()
  df_copy = df_copy.transpose()
  tmp_sum = df_copy.sum(axis=0)
  tmp_sum = tmp_sum.abs()
  tmp_sum.sort_values(inplace=True, ascending=False)

  tmp_sum = tmp_sum[tmp_sum > threshold]
  keep_rows = sorted(tmp_sum.index.values.tolist())

  if len(keep_rows) < len(ini_rows):
    df = grab_df_subset(df, keep_rows=keep_rows)

  return df

def df_filter_col_sum(df, threshold, take_abs=True):
  ''' filter columns in matrix at some threshold
  and remove rows that have all zero values '''

  from copy import deepcopy
  from .__init__ import Network
  net = Network()

  if take_abs is True:
    df_copy = deepcopy(df.abs())
  else:
    df_copy = deepcopy(df)

  df_copy = df_copy.transpose()
  df_copy = df_copy[df_copy.sum(axis=1) > threshold]
  df_copy = df_copy.transpose()
  df_copy = df_copy[df_copy.sum(axis=1) > 0]

  if take_abs is True:
    inst_rows = df_copy.index.tolist()
    inst_cols = df_copy.columns.tolist()
    df = grab_df_subset(df, inst_rows, inst_cols)

  else:
    df = df_copy

  return df

def grab_df_subset(df, keep_rows='all', keep_cols='all'):
  if keep_cols != 'all':
    df = df[keep_cols]
  if keep_rows != 'all':
    df = df.loc[keep_rows]
  return df

def get_sorted_rows(df, rank_type='sum'):
  from copy import deepcopy

  inst_df = deepcopy(df)
  inst_df = inst_df.transpose()

  if rank_type == 'sum':
    tmp_sum = inst_df.sum(axis=0)
  elif rank_type == 'var':
    tmp_sum = inst_df.var(axis=0)

  tmp_sum = tmp_sum.abs()
  tmp_sum.sort_values(inplace=True, ascending=False)
  rows_sorted = tmp_sum.index.values.tolist()

  return rows_sorted

def filter_N_top(inst_rc, df, N_top, rank_type='sum'):

  if inst_rc == 'col':
    for inst_type in df:
      df[inst_type] = df[inst_type].transpose()

  rows_sorted = get_sorted_rows(df, rank_type)

  keep_rows = rows_sorted[:N_top]

  df = df.loc[keep_rows]

  if inst_rc == 'col':
    for inst_type in df:
      df[inst_type] = df[inst_type].transpose()

  return df

def filter_threshold(df, inst_rc, threshold, num_occur=1):
  '''
  Filter a network's rows or cols based on num_occur values being above a
  threshold (in absolute_value)
  '''
  from copy import deepcopy

  inst_df = deepcopy(df)

  if inst_rc == 'col':
    inst_df = inst_df.transpose()

  inst_df = inst_df.abs()

  ini_rows = inst_df.index.values.tolist()

  inst_df[inst_df < threshold] = 0
  inst_df[inst_df >= threshold] = 1

  tmp_sum = inst_df.sum(axis=1)

  tmp_sum = tmp_sum[tmp_sum >= num_occur]

  keep_names = tmp_sum.index.values.tolist()

  if inst_rc == 'row':
    if len(keep_names) < len(ini_rows):
      df = grab_df_subset(df, keep_rows=keep_names)

  elif inst_rc == 'col':
    inst_df = inst_df.transpose()

    inst_rows = inst_df.index.values.tolist()
    inst_cols = keep_names

    df = grab_df_subset(df, inst_rows, inst_cols)

  return df

def filter_cat(net, axis, cat_index, cat_name):

  try:
    df = net.export_df()

    # DataFrame filtering will be run always be run on columns if the user
    # wants to filter rows, transpose the matrix before and after
    if axis == 'row':
      df = df.transpose()

    all_names = df.columns.tolist()

    found_names = [i for i in all_names if i[cat_index] == cat_name]

    if len(found_names) > 0:
      df = df[found_names]

      if axis == 'row':
        df = df.transpose()
    else:
      print('no ' + axis + 's were found with this category and filtering was not run')

    net.load_df(df)

  except:
    print('category filtering did not run\n check that your category filtering is set up correctly')


def filter_names(net, axis, names):

  print('filter_names')
  print(names)

  try:

    df = net.export_df()

    # Dataframe filtering will always be run on the columns. If the user wants to filter rows, then it will transpose back and forth.

    if axis == 'row':
      df = df.transpose()

    all_names = df.columns.tolist()

    found_names = []
    for inst_name in all_names:

      if type(inst_name) is tuple:
        check_name = inst_name[0]
      else:
        check_name = inst_name

      if ': ' in check_name:
        check_name = check_name.split(': ')[1]

      if check_name in names:
        found_names.append(inst_name)

    if len(found_names) > 0:
      df = df[found_names]

      if axis == 'row':
        df = df.transpose()

      net.load_df(df)

    else:
      print('no ' + axis + 's were found with these names')

  except:
    print('error in filtering names')

  print(found_names)


"""
Module for pre-processing to generate LandscapeFiles from ST data.
"""

try:
    import pyvips
except ImportError:
    pyvips = None

from pathlib import Path
import numpy as np
import pandas as pd
import os
import hashlib
import base64
from shapely.affinity import affine_transform
from shapely.geometry import MultiPolygon

import matplotlib.pyplot as plt
from matplotlib.colors import to_hex

import json

from .landscape import *
from .trx_tile import *
from .boundary_tile import *

def convert_long_id_to_short(df):
    """
    Converts a column of long integer cell IDs in a DataFrame to a shorter, hash-based representation.

    Args:
        df (pd.DataFrame): The DataFrame containing the EntityID.
    Returns:
        pd.DataFrame: The original DataFrame with an additional column named `cell_id`
                      containing the shortened cell IDs.

    The function applies a SHA-256 hash to each cell ID, encodes the hash using base64, and truncates
    it to create a shorter identifier that is added as a new column to the DataFrame.
    """
    # Function to hash and encode the cell ID
    def hash_and_shorten_id(cell_id):
        # Create a hash of the cell ID
        cell_id_bytes = str(cell_id).encode('utf-8')
        hash_object = hashlib.sha256(cell_id_bytes)
        hash_digest = hash_object.digest()

        # Encode the hash to a base64 string to mix letters and numbers, truncate to 9 characters
        short_id = base64.urlsafe_b64encode(hash_digest).decode('utf-8')[:9]
        return short_id

    # Apply the hash_and_shorten_id function to each cell ID in the specified column
    df['cell_id'] = df['EntityID'].apply(hash_and_shorten_id)

    return df


def reduce_image_size(image_path, scale_image=0.5, path_landscape_files=""):
    """

    Parameters
    ----------
    image_path : str
        Path to the image file
    scale_image : float (default=0.5)
        Scale factor for the image resize

    Returns
    -------
    new_image_path : str
        Path to the resized image file
    """

    image = pyvips.Image.new_from_file(image_path, access="sequential")

    resized_image = image.resize(scale_image)

    new_image_name = image_path.split("/")[-1].replace(".tif", "_downsize.tif")
    new_image_path = f"{path_landscape_files}/{new_image_name}"
    resized_image.write_to_file(new_image_path)

    return new_image_path


def convert_to_jpeg(image_path, quality=80):
    """
    Convert a TIFF image to a JPEG image with a quality of score

    Parameters
    ----------
    image_path : str
        Path to the image file
    quality : int (default=80)
        Quality score for the JPEG image

    Returns
    -------
    new_image_path : str
        Path to the JPEG image file

    """

    # Load the TIFF image
    image = pyvips.Image.new_from_file(image_path, access="sequential")

    # Save the image as a JPEG with a quality of 80
    new_image_path = image_path.replace(".tif", ".jpeg")
    image.jpegsave(new_image_path, Q=quality)

    return new_image_path

def convert_to_png(image_path):
    """
    Convert a TIFF image to a JPEG image with a quality of score

    Parameters
    ----------
    image_path : str
        Path to the image file
    quality : int (default=80)
        Quality score for the JPEG image

    Returns
    -------
    new_image_path : str
        Path to the JPEG image file

    """

    # Load the TIFF image
    image = pyvips.Image.new_from_file(image_path, access="sequential")

    # Save the image as a JPEG with a quality of 80
    new_image_path = image_path.replace(".tif", ".png")
    image.pngsave(new_image_path)

    return new_image_path



def convert_to_webp(image_path, quality=100):
    """
    Convert a TIFF image to a WEBP image with a specified quality score.

    Parameters
    ----------
    image_path : str
        Path to the image file
    quality : int (default=100)
        Quality score for the WEBP image (higher is better quality)

    Returns
    -------
    new_image_path : str
        Path to the WEBP image file
    """
    # Load the TIFF image
    image = pyvips.Image.new_from_file(image_path, access="sequential")

    # Save the image as a WEBP with specified quality
    new_image_path = image_path.replace(".tif", ".webp")
    image.webpsave(new_image_path, Q=quality)

    return new_image_path



def make_deepzoom_pyramid(
    image_path, output_path, pyramid_name, tile_size=512, overlap=0, suffix=".jpeg"
):
    """
    Create a DeepZoom image pyramid from a JPEG image

    Parameters
    ----------
    image_path : str
        Path to the JPEG image file
    tile_size : int (default=512)
        Tile size for the DeepZoom pyramid
    overlap : int (default=0)
        Overlap size for the DeepZoom pyramid
    suffix : str (default='jpeg')
        Suffix for the DeepZoom pyramid tiles

    Returns
    -------
    None

    """

    # Define the output path
    output_path = Path(output_path)

    # Load the JPEG image
    image = pyvips.Image.new_from_file(image_path, access="sequential")

    # check if the output path exists and create it if it does not
    output_path.mkdir(parents=True, exist_ok=True)

    # append the pyramid name to the output path
    output_path = output_path / pyramid_name

    # Save the image as a DeepZoom image pyramid
    image.dzsave(output_path, tile_size=tile_size, overlap=overlap, suffix=suffix)


def make_meta_cell_image_coord(
    technology,
    path_transformation_matrix,
    path_meta_cell_micron,
    path_meta_cell_image,
    image_scale
):
    """
    Apply an affine transformation to the cell coordinates in microns and save
    the transformed coordinates in pixels

    Parameters
    ----------
    technology : str
        The technology used to generate the data, Xenium and MERSCOPE are supported.
    path_transformation_matrix : str
        Path to the transformation matrix file
    path_meta_cell_micron : str
        Path to the meta cell file with coordinates in microns
    path_meta_cell_image : str
        Path to save the meta cell file with coordinates in pixels

    Returns
    -------
    None

    Examples
    --------
    >>> make_meta_cell_image_coord(
    ...     technology='Xenium',
    ...     path_transformation_matrix='data/transformation_matrix.txt',
    ...     path_meta_cell_micron='data/meta_cell_micron.csv',
    ...     path_meta_cell_image='data/meta_cell_image.parquet'
    ... )

    """

    transformation_matrix = pd.read_csv(
        path_transformation_matrix, header=None, sep=" "
    ).values

    if technology == "MERSCOPE":
        meta_cell = pd.read_csv(path_meta_cell_micron, usecols=["EntityID", "center_x", "center_y"])
        meta_cell = convert_long_id_to_short(meta_cell)
        meta_cell["name"] =  meta_cell["cell_id"]
        meta_cell = meta_cell.set_index('cell_id')
    elif technology == "Xenium":
        usecols = ["cell_id", "x_centroid", "y_centroid"]
        meta_cell = pd.read_csv(path_meta_cell_micron, index_col=0, usecols=usecols)
        meta_cell.columns = ["center_x", "center_y"]
        meta_cell["name"] = pd.Series(meta_cell.index, index=meta_cell.index)

    # Adding a ones column to accommodate for affine transformation
    meta_cell["ones"] = 1

    # Preparing the data for matrix multiplication
    points = meta_cell[["center_x", "center_y", "ones"]].values

    # Applying the transformation matrix
    transformed_points = np.dot(transformation_matrix, points.T).T

    # Updating the DataFrame with transformed coordinates
    meta_cell["center_x"] = transformed_points[:, 0]
    meta_cell["center_y"] = transformed_points[:, 1]

    # Dropping the ones column as it's no longer needed
    meta_cell.drop(columns=["ones"], inplace=True)

    meta_cell["center_x"] = meta_cell["center_x"] / image_scale
    meta_cell["center_y"] = meta_cell["center_y"] / image_scale

    meta_cell["geometry"] = meta_cell.apply(
        lambda row: [row["center_x"], row["center_y"]], axis=1
    )

    if technology == "MERSCOPE":
        meta_cell = meta_cell[["name", "geometry", "EntityID"]]
    else:
        meta_cell = meta_cell[["name", "geometry"]]


    meta_cell.to_parquet(path_meta_cell_image)



def make_meta_gene(technology, path_cbg, path_output):
    """
    Create a DataFrame with genes and their assigned colors

    Parameters
    ----------
    technology : str
        The technology used to generate the data, Xenium and MERSCOPE are supported.
    path_cbg : str
        Path to the cell-by-gene matrix data (the data format can vary based on technology)
    path_output : str
        Path to save the meta gene file

    Returns
    -------
    None

    Examples
    --------
    >>> make_meta_gene(
    ...     technology='Xenium',
    ...     path_cbg='data/',
    ...     path_output='data/meta_gene.parquet'
    ... )
    """

    # check if path_cbg is a path or a DataFrame
    if isinstance(path_cbg, pd.DataFrame):
        cbg = path_cbg
        genes = cbg.columns.tolist()
    else:
        if technology == "MERSCOPE":
            cbg = pd.read_csv(path_cbg, index_col=0)
            genes = cbg.columns.tolist()
        elif technology == "Xenium":
            # genes = pd.read_csv(path_cbg + 'features.tsv.gz', sep='\t', header=None)[1].values.tolist()
            cbg = read_cbg_mtx(path_cbg)
            genes = cbg.columns.tolist()

    # Get all categorical color palettes from Matplotlib and flatten them into a single list of colors
    palettes = [plt.get_cmap(name).colors for name in plt.colormaps() if "tab" in name]
    flat_colors = [color for palette in palettes for color in palette]

    # Convert RGB tuples to hex codes
    flat_colors_hex = [to_hex(color) for color in flat_colors]

    # Use modular arithmetic to assign a color to each gene, white for genes with "Blank"
    colors = [
        flat_colors_hex[i % len(flat_colors_hex)] if "Blank" not in gene else "#FFFFFF"
        for i, gene in enumerate(genes)
    ]

    # Create a DataFrame with genes and their assigned colors
    ser_color = pd.Series(colors, index=genes)

    # calculate gene expression metadata
    meta_gene = calc_meta_gene_data(cbg)
    meta_gene['color'] = ser_color

    # Identify sparse columns
    sparse_cols = [col for col in meta_gene.columns if pd.api.types.is_sparse(meta_gene[col])]

    # Convert sparse columns to dense
    for col in sparse_cols:
        meta_gene[col] = meta_gene[col].sparse.to_dense()

    meta_gene.to_parquet(path_output)


def get_max_zoom_level(path_image_pyramid):
    """
    Returns the maximum zoom level based on the highest-numbered directory
    in the specified path_image_pyramid.

    Parameters:
        path_image_pyramid (str): The path to the directory containing zoom level directories.

    Returns:
        max_pyramid_zoom (int): The maximum zoom level.
    """
    # List all entries in the path_image_pyramid that are directories and can be converted to integers
    zoom_levels = [
        entry
        for entry in os.listdir(path_image_pyramid)
        if os.path.isdir(os.path.join(path_image_pyramid, entry)) and entry.isdigit()
    ]

    # Convert to integer and find the maximum value
    max_pyramid_zoom = max(map(int, zoom_levels)) if zoom_levels else None

    return max_pyramid_zoom


def save_landscape_parameters(
    technology, path_landscape_files, image_name="dapi_files", tile_size=1000, image_info={}, image_format='.webp'
):
    """
    Save the landscape parameters to a JSON file.
    """

    if image_info == {}:
        landscape_parameters = {
            "technology": technology,
        }
    else:

        path_image_pyramid = f"{path_landscape_files}/pyramid_images/{image_name}"

        max_pyramid_zoom = get_max_zoom_level(path_image_pyramid)

        landscape_parameters = {
            "technology": technology,
            "max_pyramid_zoom": max_pyramid_zoom,
            "tile_size": tile_size,
            "image_info": image_info,
            "image_format": image_format
        }

    path_landscape_parameters = f"{path_landscape_files}/landscape_parameters.json"

    with open(path_landscape_parameters, "w") as file:
        json.dump(landscape_parameters, file, indent=4)


__all__ = ["landscape", "trx_tile", "boundary_tile"]



import pandas as pd
import numpy as np
from copy import deepcopy

def run_norm(net, df=None, norm_type='zscore', axis='row', z_clip=None):
  '''
  A dataframe can be passed to run_norm and a normalization will be run (
  e.g. zscore) on either the rows or columns
  '''

  if df is None:
    df = net.dat_to_df()

  if norm_type == 'zscore':
    df, ser_mean, ser_std = zscore_df(df, axis, z_clip=z_clip)

    net.dat['pre_zscore'] = {}
    net.dat['pre_zscore']['mean'] = ser_mean.values.tolist()
    net.dat['pre_zscore']['std'] = ser_std.values.tolist()

  if norm_type == 'qn':
    df = qn_df(df, axis)

  if norm_type == 'umi':
    df = umi_norm(df)

  net.df_to_dat(df)

  # if norm_type == 'zscore' and axis == 'row':
  #   net.dat['pre_zscore'] = {}
  #   net.dat['pre_zscore']['mean'] = ser_mean
  #   net.dat['pre_zscore']['std'] = ser_std

def qn_df(df, axis='row'):
  '''
  do quantile normalization of a dataframe dictionary, does not write to net
  '''
  # using transpose to do row qn
  if axis == 'row':
    df = df.transpose()

  missing_values = df.isnull().values.any()

  # make mask of missing values
  if missing_values:

    # get nan mask
    missing_mask = pd.isnull(df)

    # tmp fill in na with zero, will not affect qn
    df = df.fillna(value=0)

  # calc common distribution
  common_dist = calc_common_dist(df)

  # swap in common distribution
  df = swap_in_common_dist(df, common_dist)

  # swap back in missing values
  if missing_values:
    df = df.mask(missing_mask, other=np.nan)

  # using transpose to do row qn
  if axis == 'row':
    df = df.transpose()

  df_qn = df

  return df_qn

def swap_in_common_dist(df, common_dist):

  col_names = df.columns.tolist()

  qn_arr = np.array([])
  orig_rows = df.index.tolist()

  # loop through each column
  for inst_col in col_names:

    # get the sorted list of row names for the given column
    tmp_series = deepcopy(df[inst_col])
    tmp_series = tmp_series.sort_values(ascending=False)
    sorted_names = tmp_series.index.tolist()

    qn_vect = np.array([])
    for inst_row in orig_rows:
      inst_index = sorted_names.index(inst_row)
      inst_val = common_dist[inst_index]
      qn_vect = np.hstack((qn_vect, inst_val))

    if qn_arr.shape[0] == 0:
      qn_arr = qn_vect
    else:
      qn_arr = np.vstack((qn_arr, qn_vect))

  # transpose (because of vstacking)
  qn_arr = qn_arr.transpose()

  qn_df = pd.DataFrame(data=qn_arr, columns=col_names, index=orig_rows)

  return qn_df

def calc_common_dist(df):
  '''
  calculate a common distribution (for col qn only) that will be used to qn
  '''

  # axis is col
  tmp_arr = np.array([])

  col_names = df.columns.tolist()

  for inst_col in col_names:

    # sort column
    tmp_vect = df[inst_col].sort_values(ascending=False).values

    # stacking rows vertically (will transpose)
    if tmp_arr.shape[0] == 0:
      tmp_arr = tmp_vect
    else:
      tmp_arr = np.vstack((tmp_arr, tmp_vect))

  tmp_arr = tmp_arr.transpose()

  common_dist = tmp_arr.mean(axis=1)

  return common_dist

def zscore_df(df, axis='row', z_clip=None):
  '''
  take the zscore of a dataframe dictionary, does not write to net (self)
  '''

  if axis == 'row':
    df = df.transpose()

  ser_mean = df.mean()
  ser_std = df.std()

  df_z = (df - ser_mean)/ser_std

  if axis == 'row':
    df_z = df_z.transpose()

  if z_clip is not None:
    df_z = z_clip_fun(df_z, lower=-z_clip, upper=z_clip)

  return df_z, ser_mean, ser_std

def umi_norm(df):
    # umi norm
    barcode_umi_sum = df.sum()
    df_umi = df.div(barcode_umi_sum)
    return df_umi


def z_clip_fun(df_z, lower=None, upper=None):
  '''
  Trim values at input thresholds using pandas function
  '''
  df_z = df_z.clip(lower=lower, upper=upper)

  return df_z


import os
import json
import pandas as pd
import numpy as np
import geopandas as gpd
import seaborn as sns
import matplotlib.pyplot as plt
from shapely.geometry import Polygon, MultiPolygon
from ..pre.landscape import read_cbg_mtx
from ..pre.boundary_tile import get_cell_polygons

def get_largest_polygon(geometry):
    if isinstance(geometry, MultiPolygon):
        return max(geometry.geoms, key=lambda p: p.area)
    return geometry

def qc_segmentation(base_path, path_output=None, path_meta_cell_micron=None):

    """
    Calculate segmentation quality control (QC) metrics for imaging spatial transcriptomics data.

    This function computes QC metrics to assess the quality of cell segmentation and transcript assignment
    in spatial transcriptomics datasets. Metrics include transcript assignment proportion, cell count,
    mean cell area, and transcript and gene distribution statistics.

    Parameters
    ----------
    base_path : str
        Path to the data directory

    Returns
    -------
    None
        Outputs two CSV files containing cell-level and gene-specific QC metrics.

    Example
    -------
    qc_segmentation(base_path="path/to/data")

    """

    metrics = {}

    try:
        if os.path.exists(os.path.join(base_path, "segmentation_parameters.json")):
            with open(os.path.join(base_path, "segmentation_parameters.json"), 'r') as parameter_file:
                segmentation_parameters = json.load(parameter_file)
        else:
            print("segmentation_parameters.json does not exist")
    except Exception as e:
        print(f"An error occurred: {e}")

    if segmentation_parameters['technology'] == 'custom':
        cell_index = "cell_index"
        gene = "gene"
        transcript_index = "transcript_index"

        trx = pd.read_parquet(os.path.join(base_path, "transcripts.parquet"))
        trx_meta = trx[trx[cell_index] != -1][[transcript_index, cell_index, gene]]
        cell_gdf = gpd.read_parquet(os.path.join(base_path, "cell_polygons.parquet"))
        cell_meta_gdf = gpd.read_parquet(os.path.join(base_path, "cell_metadata.parquet"))
        
    elif segmentation_parameters['technology'] == 'Xenium':
        cell_index = "cell_id"
        gene = "feature_name"
        transcript_index = "transcript_id"

        trx = pd.read_parquet(os.path.join(base_path, "transcripts.parquet"))
        trx = trx.rename(columns={'name': gene})
        trx_meta = trx[trx[cell_index] != 'UNASSIGNED'][[transcript_index, cell_index, gene]]
        
        cell_gdf = get_cell_polygons(technology=segmentation_parameters['technology'], 
                                     path_cell_boundaries=os.path.join(base_path, "cell_boundaries.parquet"))
        
        cell_gdf = gpd.GeoDataFrame(geometry=cell_gdf["geometry"])
        cell_gdf["geometry"] = cell_gdf["geometry"].apply(get_largest_polygon)

        cell_gdf.reset_index(inplace=True)
        cell_gdf['area'] = cell_gdf['geometry'].area
        cell_gdf['centroid'] = cell_gdf['geometry'].centroid
        cell_meta_gdf = cell_gdf[['cell_id', 'area', 'centroid']]

    elif segmentation_parameters['technology'] == 'MERSCOPE':
        cell_index = 'EntityID'
        gene = "gene"
        transcript_index = 'transcript_id'
        
        trx = pd.read_csv(os.path.join(base_path, "detected_transcripts.csv"))
        trx = trx.rename(columns={'name': gene})
        trx_meta = trx[trx[cell_index] != -1][[transcript_index, cell_index, gene]]

        cell_gdf = get_cell_polygons(technology=segmentation_parameters['technology'], 
                                     path_cell_boundaries=os.path.join(base_path, "cell_boundaries.parquet"), 
                                     path_output=path_output,
                                     path_meta_cell_micron=path_meta_cell_micron)
        
        cell_gdf["geometry"] = cell_gdf["Geometry"].apply(get_largest_polygon)
        cell_gdf.drop(['Geometry'], axis=1, inplace=True)
        cell_gdf = gpd.GeoDataFrame(geometry=cell_gdf["Geometry"])
        
        cell_gdf.reset_index(inplace=True)
        cell_gdf['area'] = cell_gdf['geometry'].area
        cell_gdf['centroid'] = cell_gdf['geometry'].centroid
        cell_meta_gdf = cell_gdf[['cell_id', 'area', 'centroid']]

    percentage_of_assigned_transcripts = (len(trx_meta) / len(trx))

    metrics['dataset_name'] = segmentation_parameters['dataset_name']
    metrics['segmentation_approach'] = segmentation_parameters['segmentation_approach']
    
    metrics['proportion_assigned_transcripts'] = percentage_of_assigned_transcripts
    metrics['number_cells'] = len(cell_gdf)
    metrics['mean_cell_area'] = cell_gdf['geometry'].area.mean()
    
    metrics['mean_transcripts_per_cell'] = trx_meta.groupby(cell_index).size().mean()
    metrics['median_transcripts_per_cell'] = trx_meta.groupby(cell_index)[transcript_index].count().median()

    metrics['average_genes_per_cell'] = trx_meta.groupby(cell_index)[gene].nunique().mean()
    metrics['median_genes_per_cell'] = trx_meta.groupby(cell_index)[gene].nunique().median()

    metrics['proportion_empty_cells'] = ((len(cell_meta_gdf) - len(cell_gdf)) / len(cell_meta_gdf))

    metrics_df = pd.DataFrame([metrics])
    metrics_df = metrics_df.T
    metrics_df.columns = [f"{segmentation_parameters['dataset_name']}_{segmentation_parameters['segmentation_approach']}"]
    metrics_df = metrics_df.T
    
    gene_specific_metrics_df = pd.DataFrame({
        "proportion_of_cells_expressing": (trx_meta.groupby(gene)[cell_index].nunique()) / len(cell_gdf),
        "average_expression": (trx_meta.groupby(gene)[cell_index].nunique()) / (trx_meta.groupby(gene)[cell_index].nunique().sum()),
        "assigned_transcripts": (trx_meta.groupby(gene)[transcript_index].count() / trx.groupby(gene)[transcript_index].count()).fillna(0)
    })

    metrics_df.to_csv(os.path.join(base_path, "cell_specific_qc.csv"))
    gene_specific_metrics_df.to_csv(os.path.join(base_path, "gene_specific_qc.csv"))

    print("segmentation metrics calculation completed")

def classify_cells(dataframe, cell_A_name, cell_B_name, threshold_for_A_cell_classification, threshold_for_B_cell_classification):
    dataframe['Classification'] = np.where(
        dataframe[f'Total {cell_A_name} transcripts'] >= threshold_for_A_cell_classification, cell_A_name,
        np.where(dataframe[f'Total {cell_B_name} transcripts'] >= threshold_for_B_cell_classification, cell_B_name, 'Orthogonal Expression')
    )
    return dataframe

def filter_orthogonal_expression(dataframe, cell_A_name, cell_B_name, threshold_for_orthogonal_exp):
    A_cells_with_B_genes = dataframe[
        (dataframe['Classification'] == cell_A_name) &
        (dataframe[f'Total {cell_B_name} transcripts'] > threshold_for_orthogonal_exp)
    ]
    B_cells_with_A_genes = dataframe[
        (dataframe['Classification'] == cell_B_name) &
        (dataframe[f'Total {cell_A_name} transcripts'] > threshold_for_orthogonal_exp)
    ]
    return len(A_cells_with_B_genes)/len(dataframe[f'Total {cell_A_name} transcripts']), len(B_cells_with_A_genes)/len(dataframe[f'Total {cell_B_name} transcripts'])

def orthogonal_expression_calc(base_paths, cell_type_A_specific_genes, 
                          cell_type_B_specific_genes, cell_A_name, cell_B_name, threshold_for_A_cell_classification=3, threshold_for_B_cell_classification=3, threshold_for_orthogonal_exp=3, cmap='cividis'):
    
    """
    Analyze and visualize orthogonal expression patterns of cell-type-specific genes across multiple segmentation algorithms.

    This function calculates the overlap of specific genes for two cell types (A and B) within cells across multiple segmentation algorithms. 
    It then generates a histogram comparing the total transcripts for each cell type in cells that express genes from both cell types.

    Parameters
    ----------
    base_path : str
        Path to the data directory

    cell_type_A_specific_genes : list of str
        List of genes specific to cell type A.

    cell_type_B_specific_genes : list of str
        List of genes specific to cell type B.

    cell_A_name : str
        Name or label for cell type A (used in plot labeling).

    cell_B_name : str
        Name or label for cell type B (used in plot labeling).

    threshold : int
        Threshold to perform orthogonal expression quantification.


    Returns
    -------
    None
        Displays histograms comparing total transcripts for cell types A and B, grouped by segmentation algorithm.

    Example
    -------
    
    orthogonal_expression_calc(
        base_path="path/to/data",
        cell_type_A_specific_genes=["GeneA1", "GeneA2"],
        cell_type_B_specific_genes=["GeneB1", "GeneB2"],
        cell_A_name="CellTypeA",
        cell_B_name="CellTypeB"
    )
    
    """
        
    cbg_dict = {}

    cell_A_with_B_cell_specific_genes = {}
    cell_B_with_A_cell_specific_genes = {}

    for base_path in base_paths:

        with open(os.path.join(base_path, "segmentation_parameters.json"), 'r') as parameter_file:
            segmentation_parameters = json.load(parameter_file)

            if segmentation_parameters['technology'] == 'custom':
                cbg_dict[segmentation_parameters['segmentation_approach']] = pd.read_parquet(os.path.join(base_path, 
                                                                                                    "cell_by_gene_matrix.parquet"))
            elif segmentation_parameters['technology'] == 'Xenium':
                cbg_dict[segmentation_parameters['segmentation_approach']] = read_cbg_mtx(os.path.join(base_path, "cell_feature_matrix"))
                
            elif segmentation_parameters['technology'] == 'MERSCOPE':
                cbg_dict[segmentation_parameters['segmentation_approach']] = pd.read_csv(os.path.join(base_path, 
                                                                                                    "cell_by_gene_matrix.csv"))

    for algorithm_name, cbg in cbg_dict.items():

        A_cell_overlap = [gene for gene in cell_type_A_specific_genes if gene in cbg.columns]
        B_cell_overlap = [gene for gene in cell_type_B_specific_genes if gene in cbg.columns]
        
        cells_with_A_genes = cbg[A_cell_overlap].sum(axis=1) > 0
        cells_with_B_genes = cbg[B_cell_overlap].sum(axis=1) > 0
        
        cells_with_both = cbg[cells_with_A_genes & cells_with_B_genes]
        
        A_cell_genes_expressed = cells_with_both[A_cell_overlap].apply(
            lambda row: {gene: int(row[gene]) for gene in row[row > 0].index}, axis=1
        )
        
        B_cell_genes_expressed = cells_with_both[B_cell_overlap].apply(
            lambda row: {gene: int(row[gene]) for gene in row[row > 0].index}, axis=1
        )
        
        results = pd.DataFrame({
            f"{cell_A_name} genes and transcripts": A_cell_genes_expressed,
            f"{cell_B_name} genes and transcripts": B_cell_genes_expressed
        }, index=cells_with_both.index)
        
        results[f"Total {cell_A_name} transcripts"] = A_cell_genes_expressed.apply(lambda x: sum(x.values()))
        results[f"Total {cell_B_name} transcripts"] = B_cell_genes_expressed.apply(lambda x: sum(x.values()))
        
        results["Total"] = A_cell_genes_expressed.apply(lambda x: sum(x.values())) + B_cell_genes_expressed.apply(lambda x: sum(x.values()))
        results['Technology'] = algorithm_name
        
        sns.set(style='white', rc={'figure.dpi': 250, 'axes.facecolor': (0, 0, 0, 0), 'figure.facecolor': (0, 0, 0, 0)})
    
        height_of_each_facet = 3
        aspect_ratio_of_each_facet = 1
        
        g = sns.FacetGrid(results, col="Technology", sharex=False, sharey=False,
                        margin_titles=True, despine=True, col_wrap=4,
                        height=height_of_each_facet, aspect=aspect_ratio_of_each_facet,
                        gridspec_kws={"wspace": 0.01})
        
        g.map_dataframe(
            lambda data, **kwargs: sns.histplot(
                data=data,
                x=f"Total {cell_A_name} transcripts",
                y=f"Total {cell_B_name} transcripts",
                bins=15,
                cbar=True,
                cmap=cmap,
                vmin=1,
                vmax=data[f"Total {cell_A_name} transcripts"].max(),
                **kwargs
            )
        )
        
        g.set_axis_labels(f"Total {cell_A_name} transcripts", f"Total {cell_B_name} transcripts")
        for ax in g.axes.flat:
            ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True))
            ax.yaxis.set_major_locator(plt.MaxNLocator(integer=True))
            ax.tick_params(axis='both', which='major', labelsize=8)
        
        plt.tight_layout()
        plt.show()

        results = classify_cells(results, cell_A_name, cell_B_name, threshold_for_A_cell_classification, threshold_for_B_cell_classification)
        cell_A_with_B_cell_specific_genes[algorithm_name], cell_B_with_A_cell_specific_genes[algorithm_name] = filter_orthogonal_expression(results, cell_A_name, cell_B_name, threshold_for_orthogonal_exp)

    orthogonal_data = pd.DataFrame({
        'Technology': [i for i in cell_A_with_B_cell_specific_genes.keys() for _ in range(2)],
        'Category': [f'{cell_A_name} with {cell_B_name} genes', f'{cell_B_name} with {cell_A_name} genes'] * 4,
        'Count': [gene for pair in zip(cell_A_with_B_cell_specific_genes.values(), 
                                        cell_B_with_A_cell_specific_genes.values()) 
                   for gene in pair]
    })

    fig, ax = plt.subplots(figsize=(10, 6))

    sns.barplot(data=orthogonal_data, x='Technology', y='Count', hue='Category', ax=ax)

    ax.set_title(f'Orthogonal Expression: Classified {cell_A_name} and {cell_B_name} Expressing Opposite Gene Type', fontsize=15)
    ax.set_xlabel('Technology', fontsize=15, labelpad=10)
    ax.set_ylabel('Proportion of Cells', fontsize=15, labelpad=10)
    plt.yticks(fontsize=15)
    plt.xticks(fontsize=15)

    ax.legend(title='Category', title_fontsize=15, bbox_to_anchor=(1.05, 1), loc='upper left', facecolor="white", edgecolor="black", fontsize=15)

    plt.tight_layout()
    plt.show()